# The MIT License (MIT)
#
# Copyright (c) 2018 Ryan Gibson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from math import ceil
from time import time

import numpy as np

byte_depth_to_dtype = {1: np.uint8, 2: np.uint16, 4: np.uint32, 8: np.uint64}


def roundup(x, base=1):
    return int(ceil(x / base)) * base


def lsb_interleave_bytes(carrier, payload, num_lsb, truncate=False, byte_depth=1):
    """
    Interleave the bytes of payload into the num_lsb LSBs of carrier.

    :param carrier: carrier bytes
    :param payload: payload bytes
    :param num_lsb: number of least significant bits to use
    :param truncate: if True, will only return the interleaved part
    :param byte_depth: byte depth of carrier values
    :return: The interleaved bytes
    """

    plen = len(payload)
    payload_bits = np.zeros(shape=(plen, 8), dtype=np.uint8)
    payload_bits[:plen, :] = np.unpackbits(
        np.frombuffer(payload, dtype=np.uint8, count=plen)
    ).reshape(plen, 8)

    bit_height = roundup(plen * 8 / num_lsb)
    payload_bits.resize(bit_height * num_lsb)

    carrier_dtype = byte_depth_to_dtype[byte_depth]
    carrier_bits = np.unpackbits(
        np.frombuffer(carrier, dtype=carrier_dtype, count=bit_height).view(np.uint8)
    ).reshape(bit_height, 8 * byte_depth)

    carrier_bits[:, 8 * byte_depth - num_lsb: 8 * byte_depth] = payload_bits.reshape(
        bit_height, num_lsb
    )

    ret = np.packbits(carrier_bits).tobytes()
    return ret if truncate else ret + carrier[byte_depth * bit_height:]


def lsb_deinterleave_bytes(carrier, num_bits, num_lsb, byte_depth=1):
    """
    Deinterleave num_bits bits from the num_lsb LSBs of carrier.

    :param carrier: carrier bytes
    :param num_bits: number of num_bits to retrieve
    :param num_lsb: number of least significant bits to use
    :param byte_depth: byte depth of carrier values
    :return: The deinterleaved bytes
    """

    plen = roundup(num_bits / num_lsb)
    carrier_dtype = byte_depth_to_dtype[byte_depth]
    payload_bits = np.unpackbits(
        np.frombuffer(carrier, dtype=carrier_dtype, count=plen).view(np.uint8)
    ).reshape(plen, 8 * byte_depth)[:, 8 * byte_depth - num_lsb: 8 * byte_depth]
    return np.packbits(payload_bits).tobytes()[: num_bits // 8]


def lsb_interleave_list(carrier, payload, num_lsb):
    """Runs lsb_interleave_bytes with a List[uint8] carrier.

    This is slower than working with bytes directly, but is often
    unavoidable if working with libraries that require using lists."""
    bit_height = roundup(8 * len(payload) / num_lsb)
    carrier_bytes = np.array(carrier[:bit_height], dtype=np.uint8).tobytes()
    interleaved = lsb_interleave_bytes(carrier_bytes, payload, num_lsb, truncate=True)
    carrier[:bit_height] = np.frombuffer(interleaved, dtype=np.uint8).tolist()
    return carrier


def lsb_deinterleave_list(carrier, num_bits, num_lsb):
    """Runs lsb_deinterleave_bytes with a List[uint8] carrier.

    This is slower than working with bytes directly, but is often
    unavoidable if working with libraries that require using lists."""

    plen = roundup(num_bits / num_lsb)
    carrier_bytes = np.array(carrier[:plen], dtype=np.uint8).tobytes()
    deinterleaved = lsb_deinterleave_bytes(carrier_bytes, num_bits, num_lsb)
    return deinterleaved


def test(carrier_len=10 ** 7, payload_len=10 ** 6):
    """Runs consistency tests with a random carrier and payload of byte
    lengths carrier_len and payload_len, respectively."""

    def print_results(e_rates, d_rates):
        print("\n" + "-" * 40)
        row_fmt = "| {:<7}| {:<13}| {:<13}|"
        print(row_fmt.format("# LSBs", "Encode Rate", "Decode rate"))
        for n, e, d in zip(range(1, 9), e_rates[1:], d_rates[1:]):
            print(row_fmt.format(n, e, d))
        print("-" * 40)

    current_progress = 0

    def progress():
        nonlocal current_progress
        print(
            "\rProgress: ["
            + "#" * current_progress
            + "-" * (32 - current_progress)
            + "]",
            end="",
            flush=True,
        )
        current_progress += 1

    print(
        "Testing {:.1f} MB payload -> {:.1f} MB carrier...".format(
            payload_len / 1e6, carrier_len / 1e6
        )
    )
    progress()

    carrier = os.urandom(carrier_len)
    payload = os.urandom(payload_len)
    encode_rates = [""] * 9
    decode_rates = [""] * 9

    for num_lsb in range(1, 9):
        # LSB interleavings that match carrier length
        encoded = lsb_interleave_bytes(carrier, payload, num_lsb)
        progress()
        decoded = lsb_deinterleave_bytes(encoded, 8 * payload_len, num_lsb)
        progress()

        # truncated LSB interleavings
        encode_time = time()
        truncated_encode = lsb_interleave_bytes(
            carrier, payload, num_lsb, truncate=True
        )
        encode_time = time() - encode_time
        progress()
        decode_time = time()
        truncated_decode = lsb_deinterleave_bytes(
            truncated_encode, 8 * payload_len, num_lsb
        )
        decode_time = time() - decode_time
        progress()

        encode_rates[num_lsb] = "{:<6.1f} MB/s".format(
            (payload_len / 1e6) / encode_time
        )
        decode_rates[num_lsb] = "{:<6.1f} MB/s".format(
            (payload_len / 1e6) / decode_time
        )

        if decoded != payload or truncated_decode != payload:
            print("\nTest failed at {} LSBs!".format(num_lsb))
            return False

    print_results(encode_rates, decode_rates)
    return True


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
"""
    stego_lsb.WavSteg
    ~~~~~~~~~~~~~~~~~

    This module contains functions for hiding and retrieving
    data from .wav files.

    :copyright: (c) 2015 by Ryan Gibson, see AUTHORS.md for more details.
    :license: MIT License, see LICENSE.md for more details.
"""
import logging
import math
import os
import wave
from time import time

from stego_lsb.bit_manipulation import lsb_deinterleave_bytes, lsb_interleave_bytes

log = logging.getLogger(__name__)


def hide_data(sound_path, file_path, output_path, num_lsb):
    """Hide data from the file at file_path in the sound file at sound_path"""
    if sound_path is None:
        raise ValueError("WavSteg hiding requires an input sound file path")
    if file_path is None:
        raise ValueError("WavSteg hiding requires a secret file path")
    if output_path is None:
        raise ValueError("WavSteg hiding requires an output sound file path")

    sound = wave.open(sound_path, "r")

    params = sound.getparams()
    num_channels = sound.getnchannels()
    sample_width = sound.getsampwidth()
    num_frames = sound.getnframes()
    num_samples = num_frames * num_channels

    # We can hide up to num_lsb bits in each sample of the sound file
    max_bytes_to_hide = (num_samples * num_lsb) // 8
    file_size = os.stat(file_path).st_size

    log.debug(f"Using {num_lsb} LSBs, we can hide {max_bytes_to_hide} bytes")

    start = time()
    sound_frames = sound.readframes(num_frames)
    with open(file_path, "rb") as file:
        data = file.read()
    log.debug("Files read".ljust(30) + f" in {time() - start:.2f}s")

    if file_size > max_bytes_to_hide:
        required_lsb = math.ceil(file_size * 8 / num_samples)
        raise ValueError(
            "Input file too large to hide, "
            f"requires {required_lsb} LSBs, using {num_lsb}"
        )

    if sample_width != 1 and sample_width != 2:
        # Python's wave module doesn't support higher sample widths
        raise ValueError("File has an unsupported bit-depth")

    start = time()
    sound_frames = lsb_interleave_bytes(
        sound_frames, data, num_lsb, byte_depth=sample_width
    )
    log.debug(f"{file_size} bytes hidden".ljust(30) + f" in {time() - start:.2f}s")

    start = time()
    sound_steg = wave.open(output_path, "w")
    sound_steg.setparams(params)
    sound_steg.writeframes(sound_frames)
    sound_steg.close()
    log.debug("Output wav written".ljust(30) + f" in {time() - start:.2f}s")


def recover_data(sound_path, output_path, num_lsb, bytes_to_recover):
    """Recover data from the file at sound_path to the file at output_path"""
    if sound_path is None:
        raise ValueError("WavSteg recovery requires an input sound file path")
    if output_path is None:
        raise ValueError("WavSteg recovery requires an output file path")
    if bytes_to_recover is None:
        raise ValueError("WavSteg recovery requires the number of bytes to recover")

    start = time()
    sound = wave.open(sound_path, "r")

    # num_channels = sound.getnchannels()
    sample_width = sound.getsampwidth()
    num_frames = sound.getnframes()
    sound_frames = sound.readframes(num_frames)
    log.debug("Files read".ljust(30) + f" in {time() - start:.2f}s")

    if sample_width != 1 and sample_width != 2:
        # Python's wave module doesn't support higher sample widths
        raise ValueError("File has an unsupported bit-depth")

    start = time()
    data = lsb_deinterleave_bytes(
        sound_frames, 8 * bytes_to_recover, num_lsb, byte_depth=sample_width
    )
    log.debug(
        f"Recovered {bytes_to_recover} bytes".ljust(30) + f" in {time() - start:.2f}s"
    )

    start = time()
    output_file = open(output_path, "wb+")
    output_file.write(bytes(data))
    output_file.close()
    log.debug("Written output file".ljust(30) + f" in {time() - start:.2f}s")



# -*- coding: utf-8 -*-
"""
    stego_lsb.cli
    ~~~~~~~~~~~~~

    This module provides the command line interface for:
        - hiding and recovering data in .wav files
        - hiding and recovering data in bitmap (.bmp and .png)
          files
        - detecting images which have modified using the
          LSB methods.

    :copyright: (c) 2015 by Ryan Gibson, see AUTHORS.md for more details.
    :license: MIT License, see LICENSE.md for more details.
"""
import logging

import click

from stego_lsb import LSBSteg, StegDetect, WavSteg, bit_manipulation

# enable logging output
logging.basicConfig(format="%(message)s", level=logging.INFO)
log = logging.getLogger("stego_lsb")
log.setLevel(logging.DEBUG)


@click.group()
@click.version_option()
def main(args=None):
    """Console script for stegolsb."""


@main.command(context_settings=dict(max_content_width=120))
@click.option("--hide", "-h", is_flag=True, help="To hide data in an image file")
@click.option(
    "--recover", "-r", is_flag=True, help="To recover data from an image file"
)
@click.option(
    "--analyze",
    "-a",
    is_flag=True,
    default=False,
    show_default=True,
    help="Print how much data can be hidden within an image",
)
@click.option(
    "--input", "-i", "input_fp", help="Path to an bitmap (.bmp or .png) image"
)
@click.option("--secret", "-s", "secret_fp", help="Path to a file to hide in the image")
@click.option("--output", "-o", "output_fp", help="Path to an output file")
@click.option(
    "--lsb-count",
    "-n",
    default=2,
    show_default=True,
    help="How many LSBs to use",
    type=int,
)
@click.option(
    "--compression",
    "-c",
    help="1 (best speed) to 9 (smallest file size)",
    default=1,
    show_default=True,
    type=click.IntRange(1, 9),
)
@click.pass_context
def steglsb(
        ctx, hide, recover, analyze, input_fp, secret_fp, output_fp, lsb_count, compression
):
    """Hides or recovers data in and from an image"""
    try:
        if analyze:
            LSBSteg.analysis(input_fp, secret_fp, lsb_count)

        if hide:
            LSBSteg.hide_data(input_fp, secret_fp, output_fp, lsb_count, compression)
        elif recover:
            LSBSteg.recover_data(input_fp, output_fp, lsb_count)

        if not hide and not recover and not analyze:
            click.echo(ctx.get_help())
    except ValueError as e:
        log.debug(e)
        click.echo(ctx.get_help())


@main.command()
@click.option("--input", "-i", "image_path", help="Path to an image")
@click.option(
    "--lsb-count",
    "-n",
    default=2,
    show_default=2,
    type=int,
    help="How many LSBs to display",
)
@click.pass_context
def stegdetect(ctx, image_path, lsb_count):
    """Shows the n least significant bits of image"""
    if image_path:
        StegDetect.show_lsb(image_path, lsb_count)
    else:
        click.echo(ctx.get_help())


@main.command()
@click.option("--hide", "-h", is_flag=True, help="To hide data in a sound file")
@click.option("--recover", "-r", is_flag=True, help="To recover data from a sound file")
@click.option("--input", "-i", "input_fp", help="Path to a .wav file")
@click.option(
    "--secret", "-s", "secret_fp", help="Path to a file to hide in the sound file"
)
@click.option("--output", "-o", "output_fp", help="Path to an output file")
@click.option(
    "--lsb-count",
    "-n",
    default=2,
    show_default=True,
    help="How many LSBs to use",
    type=int,
)
@click.option(
    "--bytes",
    "-b",
    "num_bytes",
    help="How many bytes to recover from the sound file",
    type=int,
)
@click.pass_context
def wavsteg(ctx, hide, recover, input_fp, secret_fp, output_fp, lsb_count, num_bytes):
    """Hides or recovers data in and from a sound file"""
    try:
        if hide:
            WavSteg.hide_data(input_fp, secret_fp, output_fp, lsb_count)
        elif recover:
            WavSteg.recover_data(input_fp, output_fp, lsb_count, num_bytes)
        else:
            click.echo(ctx.get_help())
    except ValueError as e:
        log.debug(e)
        click.echo(ctx.get_help())


@main.command()
def test():
    """Runs a performance test and verifies decoding consistency"""
    bit_manipulation.test()



# -*- coding: utf-8 -*-
"""
    stego_lsb.LSBSteg
    ~~~~~~~~~~~~~~~~~

    This module contains functions for hiding and recovering
    data from bitmap (.bmp and .png) files.

    :copyright: (c) 2015 by Ryan Gibson, see AUTHORS.md for more details.
    :license: MIT License, see LICENSE.md for more details.
"""
import logging
import os
import sys
from time import time

from PIL import Image

from stego_lsb.bit_manipulation import (
    lsb_deinterleave_list,
    lsb_interleave_list,
    roundup,
)

log = logging.getLogger(__name__)


def _str_to_bytes(x, charset=sys.getdefaultencoding(), errors="strict"):
    if x is None:
        return None
    if isinstance(x, (bytes, bytearray, memoryview)):  # noqa
        return bytes(x)
    if isinstance(x, str):
        return x.encode(charset, errors)
    if isinstance(x, int):
        return str(x).encode(charset, errors)
    raise TypeError("Expected bytes")


def prepare_hide(input_image_path, input_file_path):
    """Prepare files for reading and writing for hiding data."""
    image = Image.open(input_image_path)
    input_file = open(input_file_path, "rb")
    return image, input_file


def prepare_recover(steg_image_path, output_file_path):
    """Prepare files for reading and writing for recovering data."""
    steg_image = Image.open(steg_image_path)
    output_file = open(output_file_path, "wb+")
    return steg_image, output_file


def get_filesize(path):
    """Returns the file size in bytes of the file at path"""
    return os.stat(path).st_size


def max_bits_to_hide(image, num_lsb):
    """Returns the number of bits we're able to hide in the image using
    num_lsb least significant bits."""
    # 3 color channels per pixel, num_lsb bits per color channel.
    return int(3 * image.size[0] * image.size[1] * num_lsb)


def bytes_in_max_file_size(image, num_lsb):
    """Returns the number of bits needed to store the size of the file."""
    return roundup(max_bits_to_hide(image, num_lsb).bit_length() / 8)


def hide_message_in_image(input_image, message, num_lsb):
    """Hides the message in the input image and returns the modified
    image object.
    """
    start = time()
    # in some cases the image might already be opened
    if isinstance(input_image, Image.Image):
        image = input_image
    else:
        image = Image.open(input_image)

    num_channels = len(image.getdata()[0])
    flattened_color_data = [v for t in image.getdata() for v in t]

    # We add the size of the input file to the beginning of the payload.
    message_size = len(message)
    file_size_tag = message_size.to_bytes(
        bytes_in_max_file_size(image, num_lsb), byteorder=sys.byteorder
    )
    data = file_size_tag + _str_to_bytes(message)
    log.debug("Files read".ljust(30) + f" in {time() - start:.2f}s")

    if 8 * len(data) > max_bits_to_hide(image, num_lsb):
        raise ValueError(
            f"Only able to hide {max_bits_to_hide(image, num_lsb) // 8} bytes "
            + f"in this image with {num_lsb} LSBs, but {len(data)} bytes were requested"
        )

    start = time()
    flattened_color_data = lsb_interleave_list(flattened_color_data, data, num_lsb)
    log.debug(f"{message_size} bytes hidden".ljust(30) + f" in {time() - start:.2f}s")

    start = time()
    # PIL expects a sequence of tuples, one per pixel
    image.putdata(list(zip(*[iter(flattened_color_data)] * num_channels)))
    log.debug("Image overwritten".ljust(30) + f" in {time() - start:.2f}s")
    return image


def hide_data(
        input_image_path, input_file_path, steg_image_path, num_lsb, compression_level
):
    """Hides the data from the input file in the input image."""
    if input_image_path is None:
        raise ValueError("LSBSteg hiding requires an input image file path")
    if input_file_path is None:
        raise ValueError("LSBSteg hiding requires a secret file path")
    if steg_image_path is None:
        raise ValueError("LSBSteg hiding requires an output image file path")

    image, input_file = prepare_hide(input_image_path, input_file_path)
    image = hide_message_in_image(image, input_file.read(), num_lsb)
    input_file.close()
    image.save(steg_image_path, compress_level=compression_level)


def recover_message_from_image(input_image, num_lsb):
    """Returns the message from the steganographed image"""
    start = time()
    if isinstance(input_image, Image.Image):
        steg_image = input_image
    else:
        steg_image = Image.open(input_image)

    color_data = [v for t in steg_image.getdata() for v in t]

    file_size_tag_size = bytes_in_max_file_size(steg_image, num_lsb)
    tag_bit_height = roundup(8 * file_size_tag_size / num_lsb)

    bytes_to_recover = int.from_bytes(
        lsb_deinterleave_list(
            color_data[:tag_bit_height], 8 * file_size_tag_size, num_lsb
        ),
        byteorder=sys.byteorder,
    )

    maximum_bytes_in_image = (
            max_bits_to_hide(steg_image, num_lsb) // 8 - file_size_tag_size
    )
    if bytes_to_recover > maximum_bytes_in_image:
        raise ValueError(
            "This image appears to be corrupted.\n"
            + f"It claims to hold {bytes_to_recover} B, "
            + f"but can only hold {maximum_bytes_in_image} B with {num_lsb} LSBs"
        )

    log.debug("Files read".ljust(30) + f" in {time() - start:.2f}s")

    start = time()
    data = lsb_deinterleave_list(
        color_data, 8 * (bytes_to_recover + file_size_tag_size), num_lsb
    )[file_size_tag_size:]
    log.debug(
        f"{bytes_to_recover} bytes recovered".ljust(30) + f" in {time() - start:.2f}s"
    )
    return data


def recover_data(steg_image_path, output_file_path, num_lsb):
    """Writes the data from the steganographed image to the output file"""
    if steg_image_path is None:
        raise ValueError("LSBSteg recovery requires an input image file path")
    if output_file_path is None:
        raise ValueError("LSBSteg recovery requires an output file path")

    steg_image, output_file = prepare_recover(steg_image_path, output_file_path)
    data = recover_message_from_image(steg_image, num_lsb)
    start = time()
    output_file.write(data)
    output_file.close()
    log.debug("Output file written".ljust(30) + f" in {time() - start:.2f}s")


def analysis(image_file_path, input_file_path, num_lsb):
    """Print how much data we can hide and the size of the data to be hidden"""
    if image_file_path is None:
        raise ValueError("LSBSteg analysis requires an input image file path")

    image = Image.open(image_file_path)
    print(
        f"Image resolution: ({image.size[0]}, {image.size[1]})\n"
        + f"Using {num_lsb} LSBs, we can hide:".ljust(30)
        + f" {max_bits_to_hide(image, num_lsb) // 8} B"
    )

    if input_file_path is not None:
        print(
            "Size of input file:".ljust(30) + f" {get_filesize(input_file_path)} B\n"
        )

    print("File size tag:".ljust(30) + f" {bytes_in_max_file_size(image, num_lsb)} B")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = ["Pillow>=5.3.0", "numpy>=1.15.4", "Click>=7.0"]


setup(
    author="Ryan Gibson",
    author_email="ryanalexandergibson@gmail.com",
    name="stego_lsb",
    version="1.3.1",
    description="stego lsb",
    keywords="stego lsb",
    license="MIT",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/ragibson/Steganography",
    install_requires=requirements,
    entry_points="""
        [console_scripts]
        stegolsb=stego_lsb.cli:main
    """,
    include_package_data=True,
    packages=find_packages(include=["stego_lsb"]),
    zip_safe=False,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
)
